"""IO elements for ucal_client."""
from enum import Enum
from collections import namedtuple
import json

import numpy as np


class UcalClientException(Exception):
    """Error raised during the client usage."""


class UcalBlock:
    """
    Device read/write configuration description for a specific time period.

    Plan or program in UcalClient(and Server) is stored as a
    sequence of independent blocks(UcalBlock), where each describes what
    voltage should be applied to the sample and how data should be measured.

    :param voltage_0: list of int values in millivolts, that should be applied
      sequentially to the first heater. Pattern repeats once it reaches end.
      If set to None, no voltage is applied.
    :param voltage_1: list of int values in millivolts, that should be applied
      sequentially to the second heater. Pattern repeats once it reached end.
      If set to None, no voltage is applied.
    :param read_step_tu: read step in TimeUnit (ms by default, see UcalConfig);
      Specifies how frequently should signal be measured and saved. Must be nonzero.
    :param write_step_tu: write step in TimeUnit (ms by defaul, see UcalConfig);
      Specifies how frequently voltage from voltage_0 and voltage_1
    :param block_len_tu: duration of the UcalBlock. If set to zero, user command
      is awaited to end the block (runs infinitely).
    """

    __slots__ = [
        "read_step_tu",
        "write_step_tu",
        "block_len_tu",
        "voltage_0",
        "voltage_1"
    ]
    _VOLTAGE_MAX = 10000
    _VOLTAGE_MIN = -10000

    def __init__(self, read_step_tu, write_step_tu, block_len_tu,
                 voltage_0, voltage_1):
        """
        :param read_step_tu: int, time between adc signal measurements
            in time units.
        :param write_step_tu: int, time between dac control update in
            time units.
        :param block_len_tu: int, total duration of block in time
            units. Zero means infinite block.
        :param voltage_0: None or Iterable[int], millivolts array pattern that
            would be repeated at channel 0.
        :param voltage_1: None or Iterable[int], millivolts array pattern that
            would be repeated at channel 1.
        """
        self.read_step_tu = int(read_step_tu)
        self.write_step_tu = int(write_step_tu)
        self.block_len_tu = int(block_len_tu)
        self.voltage_0 = voltage_0
        self.voltage_1 = voltage_1
        if self.voltage_0 is not None:
            self.voltage_0 = np.array(voltage_0).astype(np.int32)
        if self.voltage_1 is not None:
            self.voltage_1 = np.array(voltage_1).astype(np.int32)
        self._validate()

    def _validate(self):
        """Validate UcalBlock attributes."""
        try:
            # all time_step params must be given and must be non-negative int
            for attr in ["read_step_tu", "write_step_tu", "block_len_tu"]:
                val = getattr(self, attr)
                assert isinstance(val, int), "'{}' must be int".format(attr)
                assert val >= 0, "'{}' must be non-negative".format(attr)
            for attr in ["voltage_0", "voltage_1"]:
                val = getattr(self, attr)
                if val is not None:
                    msg = "Use None to turn off '{}' instead of {}".format(attr, val)
                    assert len(val) != 0, msg
                    val = np.array(val)
                    setattr(self, attr, val)
                    assert np.max(val) <= self._VOLTAGE_MAX, \
                        "Voltages must be less than {}".format(self._VOLTAGE_MAX)
                    assert np.min(val) >= self._VOLTAGE_MIN, \
                        "Voltages must be higher than {}".format(self._VOLTAGE_MIN)
            if (self.voltage_0 is not None) and (self.voltage_1 is not None):
                msg = "voltage_0 and voltage_1 arrays must have same len if given"
                assert isinstance(self.voltage_0, np.ndarray), msg
                assert isinstance(self.voltage_1, np.ndarray), msg
                assert len(self.voltage_0) == len(self.voltage_1), msg
            if (self.voltage_0 is not None) or (self.voltage_0 is not None):
                msg = "write_step_tu must be non-zero when voltage is given"
                assert self.write_step_tu > 0
        except (AssertionError, TypeError) as exc:
            msg = "Invalid block configuration: {}".format(str(exc))
            raise UcalClientException(msg) from None

    def __repr__(self):
        cnum = 10
        dump = lambda x: x if len(x) < cnum else x[:cnum] + '...'
        volts_repr = "[{}; {}]".format(
            dump(str(self.voltage_0)),
            dump(str(self.voltage_1))
        )
        return "<UcalBlock:read={};write={};len={};volts={}>".format(
            self.read_step_tu, self.write_step_tu, self.block_len_tu,
            volts_repr
        )


class UcalState(Enum):
    """
    State of the Server.

    :param EXECUTING: Server is executing Block.
    :param HAVE_PLAN: Server is ready to execute Blocks
    :param NO_PLAN: Server is waiting for Blocks, user can configure device.
    :param ERROR: Server is in emergency state which needs manual fix.
    """

    EXECUTING = 'Executing'
    HAS_PLAN = 'HasPlan'
    NO_PLAN = 'NoPlan'
    ERROR = 'Error'


_UcalConfig = namedtuple(
    "UcalConfig_", [
        "board_id",
        "time_unit_size",
        "storage_frame_size"
    ]
)


class UcalConfig(_UcalConfig):
    """
    Configuration of the Server.

    :param board_id: str, name that used to init daqboard.
    :param time_unit_size: float, multiplier in relevance
        to millisecond, e.g. 10 or 0.01. All time-related values
        in block are multiplied respectively
    :param storage_frame_size: int, size of the frame in server.
        Zero means size may vary. Non-zero frame size may cause
        time measurement errors caused by resampling.
    """

    def to_message(self):
        """Serialize data in json string for server."""
        assert self.time_unit_size >= 0
        try:
            data = {
                "BoardId": str(self.board_id),
                "TimeUnitSize": float(self.time_unit_size),
                "StorageFrameSize": int(self.storage_frame_size)
            }
        except ValueError as exc:
            raise UcalClientException(str(exc))
        return json.dumps(data)

    @classmethod
    def from_message(cls, message):
        """
        Deserialize data from json string from server.

        :param message: str, serialized to json message.
        """
        data = json.loads(message)
        board_id = str(data["BoardId"])
        time_unit_size = float(data["TimeUnitSize"])
        storage_frame_size = int(data["StorageFrameSize"])
        return cls(board_id, time_unit_size, storage_frame_size)


_UcalTs = namedtuple("UcalTs", ["step", "count"])


class UcalTs(_UcalTs):
    """
    Timestamp of data from server.
    Described as the *step* (time between adjacent data points) and
    the *count* (number of steps from start time moment).
    """



"""
This package provides client for Ucal Manager Server, open source software for
nanocalorimetry - https://bitbucket.org/zimka_b/ucal_manager/.

Two most important parts of this package are UcalClient and UcalBlock.
Their docs describe the model of interaction between client and server.
"""
from .base import UcalBlock, UcalState, UcalConfig, UcalTs
from .client import UcalClient



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
# pylint: disable-all
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from ucal_client._internal_grpc import server_pb2 as server__pb2


class ServerStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetConfig = channel.unary_unary(
        '/Server/GetConfig',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=server__pb2.JsonMsg.FromString,
        )
    self.GetPlan = channel.unary_stream(
        '/Server/GetPlan',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=server__pb2.BlockMsg.FromString,
        )
    self.GetData = channel.unary_stream(
        '/Server/GetData',
        request_serializer=server__pb2.TimeStampMsg.SerializeToString,
        response_deserializer=server__pb2.FrameMsg.FromString,
        )
    self.GetState = channel.unary_unary(
        '/Server/GetState',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=server__pb2.StateMsg.FromString,
        )
    self.SetPlan = channel.stream_unary(
        '/Server/SetPlan',
        request_serializer=server__pb2.BlockMsg.SerializeToString,
        response_deserializer=server__pb2.OkMsg.FromString,
        )
    self.SetConfig = channel.unary_unary(
        '/Server/SetConfig',
        request_serializer=server__pb2.JsonMsg.SerializeToString,
        response_deserializer=server__pb2.OkMsg.FromString,
        )
    self.RunNext = channel.unary_unary(
        '/Server/RunNext',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=server__pb2.OkMsg.FromString,
        )
    self.Stop = channel.unary_unary(
        '/Server/Stop',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=server__pb2.OkMsg.FromString,
        )
    self.GetLogs = channel.unary_unary(
        '/Server/GetLogs',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=server__pb2.LogMsg.FromString,
        )
    self.DropLogs = channel.unary_unary(
        '/Server/DropLogs',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=server__pb2.OkMsg.FromString,
        )


class ServerServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def GetConfig(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetPlan(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetData(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetState(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SetPlan(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SetConfig(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RunNext(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Stop(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetLogs(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DropLogs(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ServerServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetConfig': grpc.unary_unary_rpc_method_handler(
          servicer.GetConfig,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=server__pb2.JsonMsg.SerializeToString,
      ),
      'GetPlan': grpc.unary_stream_rpc_method_handler(
          servicer.GetPlan,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=server__pb2.BlockMsg.SerializeToString,
      ),
      'GetData': grpc.unary_stream_rpc_method_handler(
          servicer.GetData,
          request_deserializer=server__pb2.TimeStampMsg.FromString,
          response_serializer=server__pb2.FrameMsg.SerializeToString,
      ),
      'GetState': grpc.unary_unary_rpc_method_handler(
          servicer.GetState,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=server__pb2.StateMsg.SerializeToString,
      ),
      'SetPlan': grpc.stream_unary_rpc_method_handler(
          servicer.SetPlan,
          request_deserializer=server__pb2.BlockMsg.FromString,
          response_serializer=server__pb2.OkMsg.SerializeToString,
      ),
      'SetConfig': grpc.unary_unary_rpc_method_handler(
          servicer.SetConfig,
          request_deserializer=server__pb2.JsonMsg.FromString,
          response_serializer=server__pb2.OkMsg.SerializeToString,
      ),
      'RunNext': grpc.unary_unary_rpc_method_handler(
          servicer.RunNext,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=server__pb2.OkMsg.SerializeToString,
      ),
      'Stop': grpc.unary_unary_rpc_method_handler(
          servicer.Stop,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=server__pb2.OkMsg.SerializeToString,
      ),
      'GetLogs': grpc.unary_unary_rpc_method_handler(
          servicer.GetLogs,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=server__pb2.LogMsg.SerializeToString,
      ),
      'DropLogs': grpc.unary_unary_rpc_method_handler(
          servicer.DropLogs,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=server__pb2.OkMsg.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'Server', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



"""All ucal client tests."""
# pylint: disable=unused-variable

import json
import pytest

from ucal_client.base import UcalBlock, UcalState, \
    UcalConfig, UcalClientException


def test_state():
    """Test UcalState."""
    assert UcalState("Error") == UcalState.ERROR
    assert UcalState("NoPlan") == UcalState.NO_PLAN
    assert UcalState("HasPlan") == UcalState.HAS_PLAN
    assert UcalState("Executing") == UcalState.EXECUTING
    assert UcalState("Executing") != UcalState.HAS_PLAN


def test_valid_config():
    """Test UcalConfig."""
    config = UcalConfig(
        board_id="DaqBoard9000",
        time_unit_size=100,
        storage_frame_size=42
    )
    config2 = UcalConfig(
        "DaqBoard9000", 100, 42
    )

    config_repr = {
        "BoardId": "DaqBoard9000",
        "TimeUnitSize": 100.,
        "StorageFrameSize": 42
    }
    assert json.loads(config.to_message()) == config_repr
    assert json.loads(config2.to_message()) == config_repr

    config3 = UcalConfig.from_message(config2.to_message())
    assert config2 == config3


@pytest.mark.xfail(raises=UcalClientException)
def test_invalid_config():
    """Test exception raise from broken config."""
    config = UcalConfig(
        board_id=42,
        time_unit_size=100,
        storage_frame_size="DaqBoard9000"
    )
    msg = config.to_message()


def test_valid_block():
    """Test valid init for UcalBlock."""
    # Simplest block.
    block = UcalBlock(
        read_step_tu=1,
        write_step_tu=2,
        block_len_tu=10,
        voltage_0=[1, 2, 3],
        voltage_1=[4, 5, 6]
    )
    # Single channel block.
    block = UcalBlock(
        read_step_tu=1,
        write_step_tu=2,
        block_len_tu=10,
        voltage_0=[1, 2, 3],
        voltage_1=None
    )
    # Single channel block.
    block = UcalBlock(
        read_step_tu=1,
        write_step_tu=2,
        block_len_tu=0,
        voltage_0=None,
        voltage_1=[1, 2, 3]
    )
    # No-voltage block with write_step_tu > 0.
    block = UcalBlock(
        read_step_tu=1,
        write_step_tu=2,
        block_len_tu=0,
        voltage_0=None,
        voltage_1=None
    )
    # No-voltage block with write_step_tu == 0.
    # (Does not matter in this case)
    block = UcalBlock(
        read_step_tu=1,
        write_step_tu=0,
        block_len_tu=0,
        voltage_0=None,
        voltage_1=None
    )


def test_invalid_block():
    """Test invalid data for UcalBlock."""
    with pytest.raises(UcalClientException):
        # Different voltage length
        block = UcalBlock(
            read_step_tu=1,
            write_step_tu=2,
            block_len_tu=10,
            voltage_0=[1, 2, 3, 4],
            voltage_1=[1, 2, 3]
        )
    with pytest.raises(UcalClientException):
        # Negative read step
        block = UcalBlock(
            read_step_tu=-2,
            write_step_tu=2,
            block_len_tu=0,
            voltage_0=None,
            voltage_1=None
        )
    with pytest.raises(UcalClientException):
        # Zero write step with voltage
        block = UcalBlock(
            read_step_tu=1,
            write_step_tu=0,
            block_len_tu=0,
            voltage_0=[9000, 0],
            voltage_1=None
        )

    with pytest.raises(UcalClientException):
        # Empty list for voltage instead of None
        block = UcalBlock(
            read_step_tu=1,
            write_step_tu=0,
            block_len_tu=0,
            voltage_0=[9000, 0],
            voltage_1=[]
        )




# -*- coding: utf-8 -*-
# pylint: disable-all
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: server.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='server.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x0cserver.proto\x1a\x1bgoogle/protobuf/empty.proto\"\x17\n\x07JsonMsg\x12\x0c\n\x04json\x18\x01 \x01(\t\"s\n\x08\x42lockMsg\x12\x15\n\rwrite_step_tu\x18\x01 \x01(\x07\x12\x14\n\x0c\x62lock_len_tu\x18\x02 \x01(\x07\x12\x14\n\x0cread_step_tu\x18\x03 \x01(\x01\x12\x11\n\tvoltage_0\x18\x04 \x03(\x11\x12\x11\n\tvoltage_1\x18\x05 \x03(\x11\"+\n\x0cTimeStampMsg\x12\x0c\n\x04step\x18\x01 \x01(\x07\x12\r\n\x05\x63ount\x18\x02 \x01(\x07\"\xb5\x01\n\x08\x46rameMsg\x12!\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32\x13.FrameMsg.DataEntry\x12\x19\n\x02ts\x18\x02 \x01(\x0b\x32\r.TimeStampMsg\x12\x0c\n\x04size\x18\x03 \x01(\x07\x1a\x1a\n\nSignalData\x12\x0c\n\x04\x64\x61ta\x18\x01 \x03(\x02\x1a\x41\n\tDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.FrameMsg.SignalData:\x02\x38\x01\"&\n\x08StateMsg\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x07\x12\x0c\n\x04name\x18\x02 \x01(\t\"\x07\n\x05OkMsg\"\x15\n\x06LogMsg\x12\x0b\n\x03row\x18\x01 \x03(\t2\xbb\x03\n\x06Server\x12/\n\tGetConfig\x12\x16.google.protobuf.Empty\x1a\x08.JsonMsg\"\x00\x12\x30\n\x07GetPlan\x12\x16.google.protobuf.Empty\x1a\t.BlockMsg\"\x00\x30\x01\x12\'\n\x07GetData\x12\r.TimeStampMsg\x1a\t.FrameMsg\"\x00\x30\x01\x12/\n\x08GetState\x12\x16.google.protobuf.Empty\x1a\t.StateMsg\"\x00\x12 \n\x07SetPlan\x12\t.BlockMsg\x1a\x06.OkMsg\"\x00(\x01\x12\x1f\n\tSetConfig\x12\x08.JsonMsg\x1a\x06.OkMsg\"\x00\x12+\n\x07RunNext\x12\x16.google.protobuf.Empty\x1a\x06.OkMsg\"\x00\x12(\n\x04Stop\x12\x16.google.protobuf.Empty\x1a\x06.OkMsg\"\x00\x12,\n\x07GetLogs\x12\x16.google.protobuf.Empty\x1a\x07.LogMsg\"\x00\x12,\n\x08\x44ropLogs\x12\x16.google.protobuf.Empty\x1a\x06.OkMsg\"\x00\x62\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,])




_JSONMSG = _descriptor.Descriptor(
  name='JsonMsg',
  full_name='JsonMsg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='json', full_name='JsonMsg.json', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=45,
  serialized_end=68,
)


_BLOCKMSG = _descriptor.Descriptor(
  name='BlockMsg',
  full_name='BlockMsg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='write_step_tu', full_name='BlockMsg.write_step_tu', index=0,
      number=1, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_len_tu', full_name='BlockMsg.block_len_tu', index=1,
      number=2, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='read_step_tu', full_name='BlockMsg.read_step_tu', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='voltage_0', full_name='BlockMsg.voltage_0', index=3,
      number=4, type=17, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='voltage_1', full_name='BlockMsg.voltage_1', index=4,
      number=5, type=17, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=70,
  serialized_end=185,
)


_TIMESTAMPMSG = _descriptor.Descriptor(
  name='TimeStampMsg',
  full_name='TimeStampMsg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='step', full_name='TimeStampMsg.step', index=0,
      number=1, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='count', full_name='TimeStampMsg.count', index=1,
      number=2, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=187,
  serialized_end=230,
)


_FRAMEMSG_SIGNALDATA = _descriptor.Descriptor(
  name='SignalData',
  full_name='FrameMsg.SignalData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='FrameMsg.SignalData.data', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=321,
  serialized_end=347,
)

_FRAMEMSG_DATAENTRY = _descriptor.Descriptor(
  name='DataEntry',
  full_name='FrameMsg.DataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='FrameMsg.DataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='FrameMsg.DataEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=349,
  serialized_end=414,
)

_FRAMEMSG = _descriptor.Descriptor(
  name='FrameMsg',
  full_name='FrameMsg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='FrameMsg.data', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ts', full_name='FrameMsg.ts', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='size', full_name='FrameMsg.size', index=2,
      number=3, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_FRAMEMSG_SIGNALDATA, _FRAMEMSG_DATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=233,
  serialized_end=414,
)


_STATEMSG = _descriptor.Descriptor(
  name='StateMsg',
  full_name='StateMsg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='StateMsg.code', index=0,
      number=1, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='StateMsg.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=416,
  serialized_end=454,
)


_OKMSG = _descriptor.Descriptor(
  name='OkMsg',
  full_name='OkMsg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=456,
  serialized_end=463,
)


_LOGMSG = _descriptor.Descriptor(
  name='LogMsg',
  full_name='LogMsg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='row', full_name='LogMsg.row', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=465,
  serialized_end=486,
)

_FRAMEMSG_SIGNALDATA.containing_type = _FRAMEMSG
_FRAMEMSG_DATAENTRY.fields_by_name['value'].message_type = _FRAMEMSG_SIGNALDATA
_FRAMEMSG_DATAENTRY.containing_type = _FRAMEMSG
_FRAMEMSG.fields_by_name['data'].message_type = _FRAMEMSG_DATAENTRY
_FRAMEMSG.fields_by_name['ts'].message_type = _TIMESTAMPMSG
DESCRIPTOR.message_types_by_name['JsonMsg'] = _JSONMSG
DESCRIPTOR.message_types_by_name['BlockMsg'] = _BLOCKMSG
DESCRIPTOR.message_types_by_name['TimeStampMsg'] = _TIMESTAMPMSG
DESCRIPTOR.message_types_by_name['FrameMsg'] = _FRAMEMSG
DESCRIPTOR.message_types_by_name['StateMsg'] = _STATEMSG
DESCRIPTOR.message_types_by_name['OkMsg'] = _OKMSG
DESCRIPTOR.message_types_by_name['LogMsg'] = _LOGMSG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

JsonMsg = _reflection.GeneratedProtocolMessageType('JsonMsg', (_message.Message,), {
  'DESCRIPTOR' : _JSONMSG,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:JsonMsg)
  })
_sym_db.RegisterMessage(JsonMsg)

BlockMsg = _reflection.GeneratedProtocolMessageType('BlockMsg', (_message.Message,), {
  'DESCRIPTOR' : _BLOCKMSG,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:BlockMsg)
  })
_sym_db.RegisterMessage(BlockMsg)

TimeStampMsg = _reflection.GeneratedProtocolMessageType('TimeStampMsg', (_message.Message,), {
  'DESCRIPTOR' : _TIMESTAMPMSG,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:TimeStampMsg)
  })
_sym_db.RegisterMessage(TimeStampMsg)

FrameMsg = _reflection.GeneratedProtocolMessageType('FrameMsg', (_message.Message,), {

  'SignalData' : _reflection.GeneratedProtocolMessageType('SignalData', (_message.Message,), {
    'DESCRIPTOR' : _FRAMEMSG_SIGNALDATA,
    '__module__' : 'server_pb2'
    # @@protoc_insertion_point(class_scope:FrameMsg.SignalData)
    })
  ,

  'DataEntry' : _reflection.GeneratedProtocolMessageType('DataEntry', (_message.Message,), {
    'DESCRIPTOR' : _FRAMEMSG_DATAENTRY,
    '__module__' : 'server_pb2'
    # @@protoc_insertion_point(class_scope:FrameMsg.DataEntry)
    })
  ,
  'DESCRIPTOR' : _FRAMEMSG,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:FrameMsg)
  })
_sym_db.RegisterMessage(FrameMsg)
_sym_db.RegisterMessage(FrameMsg.SignalData)
_sym_db.RegisterMessage(FrameMsg.DataEntry)

StateMsg = _reflection.GeneratedProtocolMessageType('StateMsg', (_message.Message,), {
  'DESCRIPTOR' : _STATEMSG,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:StateMsg)
  })
_sym_db.RegisterMessage(StateMsg)

OkMsg = _reflection.GeneratedProtocolMessageType('OkMsg', (_message.Message,), {
  'DESCRIPTOR' : _OKMSG,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:OkMsg)
  })
_sym_db.RegisterMessage(OkMsg)

LogMsg = _reflection.GeneratedProtocolMessageType('LogMsg', (_message.Message,), {
  'DESCRIPTOR' : _LOGMSG,
  '__module__' : 'server_pb2'
  # @@protoc_insertion_point(class_scope:LogMsg)
  })
_sym_db.RegisterMessage(LogMsg)


_FRAMEMSG_DATAENTRY._options = None

_SERVER = _descriptor.ServiceDescriptor(
  name='Server',
  full_name='Server',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=489,
  serialized_end=932,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetConfig',
    full_name='Server.GetConfig',
    index=0,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_JSONMSG,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetPlan',
    full_name='Server.GetPlan',
    index=1,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_BLOCKMSG,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetData',
    full_name='Server.GetData',
    index=2,
    containing_service=None,
    input_type=_TIMESTAMPMSG,
    output_type=_FRAMEMSG,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetState',
    full_name='Server.GetState',
    index=3,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_STATEMSG,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='SetPlan',
    full_name='Server.SetPlan',
    index=4,
    containing_service=None,
    input_type=_BLOCKMSG,
    output_type=_OKMSG,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='SetConfig',
    full_name='Server.SetConfig',
    index=5,
    containing_service=None,
    input_type=_JSONMSG,
    output_type=_OKMSG,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='RunNext',
    full_name='Server.RunNext',
    index=6,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_OKMSG,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='Stop',
    full_name='Server.Stop',
    index=7,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_OKMSG,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetLogs',
    full_name='Server.GetLogs',
    index=8,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_LOGMSG,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='DropLogs',
    full_name='Server.DropLogs',
    index=9,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_OKMSG,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_SERVER)

DESCRIPTOR.services_by_name['Server'] = _SERVER

# @@protoc_insertion_point(module_scope)



import os
from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


def is_requirement(line: str) -> bool:
    """
    Return True if the requirement line is a package requirement;
    that is, it is not blank, a comment, or editable.
    """
    # Remove whitespace at the start/end of the line
    line = line.strip()

    # Skip blank lines, comments, and editable installs
    return not (
        line == '' or
        line.startswith('-r') or
        line.startswith('#') or
        line.startswith('-e') or
        line.startswith('git+')
    )


def load_requirements(*requirements_paths):
    """
    Load all requirements from the specified requirements files.
    Returns a list of requirement strings.
    """
    requirements = set()
    for path in requirements_paths:
        requirements.update(
            line.strip() for line in open(path).readlines()
            if is_requirement(line)
        )
    return list(requirements)


setup(
    name='ucal_client',
    version='0.6.0',
    python_requires='>3.5.0',
    packages=find_packages(),
    description='Client for Ucal Manager Server',
    long_description=README,
    url='https://bitbucket.com/zimka/ucal_client',
    author='Boris Zimka, Igor Stepanenko',
    author_email='zimka@phystech.edu, igor.stepanenko@phystech.edu',
    install_requires=load_requirements('requirements.txt',),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering'
    ],
)



"""Ucal Client implementation."""
from functools import wraps
import warnings

import grpc

# Make deprecation warnings silent
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from google.protobuf import empty_pb2
    from ucal_client._internal_grpc import server_pb2
    from ucal_client._internal_grpc import server_pb2_grpc
    from ucal_client.base import UcalBlock, UcalState, \
        UcalConfig, UcalClientException, UcalTs

import numpy as np
import pandas as pd


_SERVER_DEFAULT_HOST = "192.168.241.1"
_SERVER_DEFAULT_PORT = "10003"
_SERVER_DEFAULT_TIMEOUT_SECONDS = 5


def grpc_reraise(method):
    """Wrap grpc error into UcalClientException."""
    @wraps(method)
    def reraised_method(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except grpc.RpcError as exc:
            exc_msg = None
            code = exc.code()  # pylint: disable=no-member
            details = exc.details()  # pylint: disable=no-member
            if code in (
                    grpc.StatusCode.UNAVAILABLE,
                    grpc.StatusCode.UNKNOWN,
                    grpc.StatusCode.DEADLINE_EXCEEDED
            ):
                exc_msg = "Failed to establish connection: {}".format(details)
            elif code == grpc.StatusCode.INVALID_ARGUMENT:
                exc_msg = "Got invalid input: {}".format(details)
            elif code == grpc.StatusCode.FAILED_PRECONDITION:
                exc_msg = "Action can't be performed: {}".format(details)
            if exc_msg:
                # No grpc traceback
                raise UcalClientException(exc_msg) from None
            else:
                # Show grpc traceback
                exc_msg = "Server bug, please report: {}".format(details)
                raise UcalClientException(exc_msg)
    return reraised_method


class UcalClient:
    """
    The main object that can be used to manage Ucal Server.
    Client (and Server) supports next actions:

    - get_state to get current state of server, see UcalState;
    - set/get_config to save global settings, see UcalConfig;
    - set/get_plan to set execution program, see UcalBlock;
    - get_data to download from server data as a pandas.DataFrame;
    - run_next/stop to start and stop program execution.
    """
    TIME_TYPES = ('expected', 'measured')
    _EXPECTED_DATA_KEYS = ('S0', 'S1', 'S2', 'S3', 'S4')
    _ADDITIONAL_KEYS = ('count', 'step')
    _TIME_KEY = 'Time'

    MERGED_DATA_KEYS = tuple([_TIME_KEY]) + _EXPECTED_DATA_KEYS
    RAW_DATA_KEYS = _ADDITIONAL_KEYS + _EXPECTED_DATA_KEYS

    def __init__(
            self,
            host=_SERVER_DEFAULT_HOST,
            port=_SERVER_DEFAULT_PORT,
            timeout_s=_SERVER_DEFAULT_TIMEOUT_SECONDS,
            use_time='expected',
            max_message_length=128*1024*1024
    ):
        """
        Set host to 'localhost' when server is run locally.
        :param host: IP addrress where server is running
        :param port: port where server is running
        :param timeout_s: how much time in seconds should
            client wait for response from server,
        :param use_time: how to associate time with points in merged data,
            either 'expected' or 'measured'.
            'expected' supposes that server device (electronic) has no time
            measurement error and uses cumulative step as measurement time
            estimation. However it does not take into account time between
            block runs, so it may accumulate systematic time measurement error.
            'measured' uses server (programmatic) global time. It has
            significant random error, which leads to non-monotonic time
            estimation, but it does not have systematic error.
        """
        self.host = host
        self.port = port
        self.timeout_s = timeout_s
        self.use_time = use_time
        if self.use_time not in self.TIME_TYPES:
            raise UcalClientException(
                "Wrong use_time value '{}', must be one of {}".format(
                    self.use_time, self.TIME_TYPES
                )
            )
        options = [
            ('grpc.max_send_message_length', max_message_length),
            ('grpc.max_receive_message_length', max_message_length),

        ]
        self._channel = grpc.insecure_channel(
            "{}:{}".format(self.host, self.port), options
        )
        self.stub = server_pb2_grpc.ServerStub(self._channel)

    def is_server_available(self, timeout_s=2):
        """
        Checks if server is currently available.
        :param timeout_s: how much time in seconds should
            client wait for response from server
        """
        try:
            grpc.channel_ready_future(self._channel).result(
                timeout=timeout_s
            )
            return True
        except grpc.FutureTimeoutError:
            return False

    @grpc_reraise
    def get_state(self):
        """
        Return UcalState of the server.
        Possible results: *NoPlan*, *HavePlan*, *Executing*, *Error*.
        State defines, which actions can or can not be executed by server now.
        """
        return UcalState(
            self.stub.GetState(empty_pb2.Empty(), timeout=self.timeout_s).name
        )

    @grpc_reraise
    def get_config(self):
        """
        Return current server configuration as a UcalConfig.
        Valid action at any state.
        """
        return UcalConfig.from_message(
            self.stub.GetConfig(empty_pb2.Empty(), timeout=self.timeout_s).json
        )

    @grpc_reraise
    def set_config(self, config):
        """
        Apply new config(UcalConfig) to the server.
        Valid action at *NoPlan* state only.

        :param config: UcalConfig, Str, None or Dict with
            valid UcalConfig key-values
        """
        if isinstance(config, dict):
            config = UcalConfig(**config)
        if isinstance(config, UcalConfig):
            self.stub.SetConfig(
                server_pb2.JsonMsg(json=config.to_message()), timeout=self.timeout_s
            )
            return
        if config is None:
            config = "{}"
        if isinstance(config, str):
            self.stub.SetConfig(
                server_pb2.JsonMsg(json=config), timeout=self.timeout_s
            )
            return
        msg = "set_config accepts dict, UcalConfig or str, got {}".format(
            type(config)
        )
        raise UcalClientException(msg)

    @grpc_reraise
    def get_plan(self):
        """
        Return plan (List[UcalBlock]) that server is executing or
        going to execute.
        Valid action at *NoPlan*, *HavePlan* and *Executing* states.
        """
        grpc_blocks = list(
            b for b in self.stub.GetPlan(empty_pb2.Empty())
        )
        replace = lambda v: v if len(v) else None
        return list(
            UcalBlock(
                read_step_tu=b.read_step_tu,
                write_step_tu=b.write_step_tu,
                block_len_tu=b.block_len_tu,
                voltage_0=replace(b.voltage_0),
                voltage_1=replace(b.voltage_1)
            ) for b in grpc_blocks
        )

    @grpc_reraise
    def set_plan(self, plan):
        """
        Set a new plan for server.
        Valid action at *NoPlan* and *HavePlan* states.

        :param plan: List[UcalBlock], starting point for returned data.
            Set empty list([]) to move to *NoPlan* state.
        """
        if not (
                isinstance(plan, list) and
                all(isinstance(b, UcalBlock) for b in plan)
        ):
            msg = "Plan must have type List[UcalBlock]"
            raise UcalClientException(msg)
        self.stub.SetPlan(
            (server_pb2.BlockMsg(
                read_step_tu=b.read_step_tu,
                write_step_tu=b.write_step_tu,
                block_len_tu=b.block_len_tu,
                voltage_0=b.voltage_0,
                voltage_1=b.voltage_1
            ) for b in plan),
            timeout=self.timeout_s
        )

    @grpc_reraise
    def get_data(self, start_ts=None, merge=True):
        """
        Return List[pd.DataFrame] from Server.
        If merge=True, DataFrames are concatenated.
        DataFrames contain info about voltage and time moments from
        the execution start.

        :param start_ts: UcalTs, starting point for returned data.
        :param merge: bool, whether to concat frames into single df.
        :param global_time: bool, whether to use global time in merged df
        """
        if start_ts is None:
            start_ts = UcalTs(0, 0)
        assert isinstance(start_ts, UcalTs)

        def parse_frame_msg(msg):
            """Turn server_pb2.FrameMsg into pd.DataFrame."""
            data = dict(msg.data)
            if len(data.keys()) and (sorted(data.keys()) != list(self._EXPECTED_DATA_KEYS)):
                msg = "Server interface has changed! "
                msg += "Expected data keys '{}' but got '{}'".format(
                    self._EXPECTED_DATA_KEYS, list(data.keys())
                )
                raise UcalClientException(msg)
            step = np.zeros(msg.size) + msg.ts.step
            count = np.arange(msg.ts.count - msg.size + 1, msg.ts.count + 1)
            return np.stack(
                [count, step] +
                [np.array(data[k].data) for k in self._EXPECTED_DATA_KEYS],
                axis=-1
            )
        raw_data = list(self.stub.GetData(
            server_pb2.TimeStampMsg(step=start_ts.step, count=start_ts.count),
            timeout=self.timeout_s
        ))
        message_arrays = list(parse_frame_msg(r) for r in raw_data)
        # drop redundant points from first frame if there are any
        if len(message_arrays):
            first_frame = message_arrays[0]
            after_ts_idx = first_frame[:, 0] > start_ts.count
            first_frame_after_ts = first_frame[after_ts_idx, :]
            message_arrays[0] = first_frame_after_ts

        if merge:
            return self._merge_data(message_arrays, start_ts)
        else:
            return message_arrays

    def _merge_data(self, message_arrays, start_ts):
        """
        Composes pd.DataFrame from raw frame arrays.
        """
        assert isinstance(message_arrays, list)
        assert all(isinstance(x, np.ndarray) for x in message_arrays)
        if start_ts is None:
            start_ts = UcalTs(0, 0)

        if not len(message_arrays):
            return pd.DataFrame(columns=self.MERGED_DATA_KEYS).set_index(self._TIME_KEY)
        total = np.vstack(message_arrays)
        tk = self._TIME_KEY
        df = pd.DataFrame(total, columns=self.RAW_DATA_KEYS)
        if self.use_time == 'measured':
            df[tk] = df['step'] * df['count']
        elif self.use_time == 'expected':
            df[tk] = df['step'].cumsum() + start_ts.count*start_ts.step
        else:
            raise UcalClientException(
                "Unexpected .use_time value '{}'".format(self.use_time)
            )
        df = df.drop(columns=['count', 'step'])
        return df.set_index(tk)

    @grpc_reraise
    def run_next(self):
        """
        Start next block execution.
        Valid action at *HavePlan*, *Executing* states.
        If server is at *HavePlan* state, execution is started.
        If server is at *Executing* state, current UcalBlock execution is
        stopped and next UcalBlock in plan is started.
        If there is no more UcalBlock in plan, execution is finished and
        *HavePlan* state is set.
        """
        return self.stub.RunNext(empty_pb2.Empty(), timeout=self.timeout_s)

    @grpc_reraise
    def stop(self):
        """
        Stop blocks execution. All measured data is available.
        Valid action at *Executing* state.
        """
        return self.stub.Stop(empty_pb2.Empty(), timeout=self.timeout_s)

    @grpc_reraise
    def get_logs(self):
        """Return current server logs."""
        return str(self.stub.GetLogs(empty_pb2.Empty(), timeout=self.timeout_s).row)

    @grpc_reraise
    def drop_logs(self):
        """Drop current server logs."""
        self.stub.DropLogs(empty_pb2.Empty(), timeout=self.timeout_s)
